import logging

from pathlib import Path
import numpy as np

from pysamosa.common_types import RetrackerBaseType, L1bSourceType, ProcMode
from pysamosa.data_access import (
    data_vars_s3,
    data_vars_s6,
    data_vars_cs,
    data_vars_dart,
    data_vars_retracker,
)
from pysamosa.retracker_processor import RetrackerProcessor
from pysamosa.settings_manager import get_default_base_settings, SettingsPreset
from pysamosa.settings import S6_DATA_DIR
from pysamosa.utils import plot_l2_results_vs_ref


def raise_wrong_sensor_type(st, st_correct):
    if st.value is not st_correct.value:
        raise RuntimeError("choose correct SensorType")


max_rmse_swh_m = 0.05
max_rmse_ssh_m = 0.01


def test_retrack_s3(s3_eum_l1b, s3_eum_l2):
    l1b, l2 = s3_eum_l1b, s3_eum_l2
    data_vars_l1b = data_vars_s3["l1b"]
    file_id = "s3_0"
    n_offset = 25480

    l1b_src_type, retracker_basetype, preset = (
        L1bSourceType.EUM_S3,
        RetrackerBaseType.SAM,
        SettingsPreset.NONE
        # L1bSourceType.EUM, RetrackerBaseType.SAM, SettingsPreset.CORALv1
    )

    # generate default settings
    (
        rp_sets,
        retrack_sets,
        fitting_sets,
        wf_sets,
        sensor_sets,
    ) = get_default_base_settings(
        retracker_basetype=retracker_basetype,
        settings_preset=preset,
        l1b_src_type=l1b_src_type,
    )

    rp_sets.n_offset = n_offset
    rp_sets.n_inds = 60
    rp_sets.n_procs = 1
    rp_sets.skip_if_exists = False

    l1b_file = l1b.get_nc_filename(file_id)

    rp = RetrackerProcessor(
        l1b_source=[l1b_file],
        l1b_data_vars=data_vars_l1b,
        rp_sets=rp_sets,
        retrack_sets=retrack_sets,
        fitting_sets=fitting_sets,
        wf_sets=wf_sets,
        sensor_sets=sensor_sets,
        log_level=logging.DEBUG,
    )
    rp.process()

    # ref datasets
    l2_ref = l2(n_offset=n_offset, n_inds=rp_sets.n_inds, file_id=file_id)

    fig, rmse_swh, rmse_ssh = plot_l2_results_vs_ref(
        rp.output_l2,
        l2_ref,
        cog_corr=0.55590,
        fig_title=f"{rp_sets.retracker_basetype}-{preset} ({l1b_src_type.value})",
    )

    assert rmse_swh < max_rmse_swh_m
    assert rmse_ssh < max_rmse_ssh_m


def test_retrack_s6(s6_eum_l1b, s6_eum_l2):
    l1b, l2 = s6_eum_l1b, s6_eum_l2
    data_vars_l1b = data_vars_s6["l1b"]
    # file_id = 's6_eum_0_f03'; n_offset = 13340;
    # file_id = 's6_eum_1_f04'; n_offset = 18600;
    # file_id = 's6_eum_1_f04'; n_offset = 34700;
    file_id = "s6_eum_2_f06"
    n_offset = 18600
    # file_id = 's6_eum_2_f06'; n_offset = 34700;

    l1b_file = l1b.get_nc_filename(file_id)

    l1b_src_type = (
        L1bSourceType.EUM_S6_F04
        if "f04" in str(l1b_file).lower()
        else L1bSourceType.EUM_S6_F06
    )
    retracker_basetype, preset = RetrackerBaseType.SAM, SettingsPreset.NONE

    # generate default settings
    (
        rp_sets,
        retrack_sets,
        fitting_sets,
        wf_sets,
        sensor_sets,
    ) = get_default_base_settings(
        retracker_basetype=retracker_basetype,
        settings_preset=preset,
        l1b_src_type=l1b_src_type,
    )

    rp_sets.n_offset = n_offset
    rp_sets.n_inds = 26
    rp_sets.n_procs = 1
    rp_sets.skip_if_exists = False

    rp = RetrackerProcessor(
        l1b_source=[l1b_file],
        l1b_data_vars=data_vars_l1b,
        rp_sets=rp_sets,
        retrack_sets=retrack_sets,
        fitting_sets=fitting_sets,
        sensor_sets=sensor_sets,
        wf_sets=wf_sets,
        log_level=logging.DEBUG,
    )

    # if 'f04' in str(l1b_file).lower():
    #     rp.retracker.model_sets.Enable_Slope_Effect_Flag = True
    # rp.retracker.model_sets.Enable_Slope_Effect_Flag = True

    rp.process()

    # ref datasets
    l2_ref = l2(n_offset=n_offset, n_inds=rp_sets.n_inds, file_id=file_id)

    fig, rmse_swh, rmse_ssh = plot_l2_results_vs_ref(
        rp.output_l2,
        l2_ref,
        fig_title=f"{rp_sets.retracker_basetype}-{preset} ({l1b_src_type.value})",
    )

    fig.show()

    assert rmse_swh < max_rmse_swh_m
    assert rmse_ssh < max_rmse_ssh_m


def test_retrack_cs(cs_eum_l1b, cs_eum_l2):
    l1b_src_type, retracker_basetype, preset = (
        L1bSourceType.EUM_CS,
        RetrackerBaseType.SAM,
        SettingsPreset.NONE,
    )
    # l1b_src_type, retracker_basetype, preset = L1bSourceType.EUM_CS, RetrackerBaseType.SAMPLUS, SettingsPreset.NONE

    l1b, l2 = cs_eum_l1b, cs_eum_l2
    data_vars_l1b = data_vars_cs["l1b"]
    # file_id = 'cs_0'; n_offset = 1000;
    file_id = "cs_1"
    n_offset = 3503
    # file_id = 'cs_2'; n_offset = 3503;

    # generate default settings
    (
        rp_sets,
        retrack_sets,
        fitting_sets,
        wf_sets,
        sensor_sets,
    ) = get_default_base_settings(
        retracker_basetype=retracker_basetype,
        settings_preset=preset,
        l1b_src_type=l1b_src_type,
    )

    rp_sets.n_offset = n_offset
    rp_sets.n_inds = 30
    rp_sets.n_procs = 1
    rp_sets.skip_if_exists = False

    l1b_file = l1b.get_nc_filename(file_id)

    rp = RetrackerProcessor(
        l1b_source=[l1b_file],
        l1b_data_vars=data_vars_l1b,
        rp_sets=rp_sets,
        retrack_sets=retrack_sets,
        fitting_sets=fitting_sets,
        sensor_sets=sensor_sets,
        wf_sets=wf_sets,
        log_level=logging.DEBUG,
    )
    rp.process()

    # ref datasets
    l2_ref = l2(n_offset=n_offset, n_inds=rp_sets.n_inds, file_id=file_id)

    fig, rmse_swh, rmse_ssh = plot_l2_results_vs_ref(
        rp.output_l2,
        l2_ref,
        fig_title=f"{rp_sets.retracker_basetype}-{preset} ({l1b_src_type.value})",
    )

    fig.show()

    # assert rmse_swh < max_rmse_swh_m
    # assert rmse_ssh < max_rmse_ssh_m


def test_retrack_ffsar(dataset_generic_l1b, dataset_generic_l2):
    proc_mode = ProcMode.FFSAR
    # proc_mode = ProcMode.UFSAR

    is_ffsar = proc_mode == ProcMode.FFSAR
    n_offset = 0
    n_inds = 20

    data_vars_l1b = (
        data_vars_dart["l1b_ffsar"]
        if proc_mode == ProcMode.FFSAR
        else data_vars_dart["l1b"]
    )
    data_vars_l2 = data_vars_retracker["l2"]
    ncfile_l1b = Path(
        S6_DATA_DIR
        / "ffsar"
        / "l1b"
        / "S6A_P4_1B_HR______20210423T162639_20210423T172054_20220514T143739_3255_016_213_106_DAR__REP_NT_F06_53.73_53.95.nc"
    )
    l1b = dataset_generic_l1b(ncfile_l1b, data_vars_l1b)

    ncfile_l2 = Path(
        S6_DATA_DIR
        / "ffsar"
        / "l2"
        / "S6A_P4_1B_HR______20210423T162639_20210423T172054_20220514T143739_3255_016_213_106_DAR__REP_NT_F06_53.73_53.95.nc"
    )
    l2 = dataset_generic_l2(ncfile_l2, data_vars_l2)

    retracker_basetype, preset = RetrackerBaseType.SAM, SettingsPreset.CORALv2
    # retracker_basetype, preset = RetrackerBaseType.SAM, SettingsPreset.NONE

    # generate default settings
    if is_ffsar:
        l1b_src_type = (
            L1bSourceType.EUM_S6_F04_FFSAR
            if "f04" in str(ncfile_l1b).lower()
            else L1bSourceType.EUM_S6_F06_FFSAR
        )
    else:
        l1b_src_type = (
            L1bSourceType.EUM_S6_F06
            if "f06" in str(ncfile_l1b).lower()
            else L1bSourceType.EUM_S6_F04
        )

    (
        rp_sets,
        retrack_sets,
        fitting_sets,
        wf_sets,
        sensor_sets,
    ) = get_default_base_settings(
        retracker_basetype=retracker_basetype,
        settings_preset=preset,
        l1b_src_type=l1b_src_type,
    )
    wf_sets.np, wf_sets.zp_oversampling_factor = (2 * 256, 2)

    rp_sets.n_offset = n_offset
    rp_sets.n_inds = n_inds
    rp_sets.n_procs = 10
    rp_sets.skip_if_exists = False
    l2_posting_rate = 20
    l1b_posting_rate = (
        int(ncfile_l1b.parent.name.split("_")[-1])
        if "coastal_ffsar" in str(ncfile_l1b)
        else 20
    )
    # rp_sets.reduce_l2_factor = 1
    rp_sets.reduce_l2_factor = round(l1b_posting_rate / l2_posting_rate)
    # rp_sets.reduce_l2_factor = 4 if '_80/' in str(ncfile_l1b) else 1
    rp_sets.dynamic_fg_epoch_n_adjacent_meas *= rp_sets.reduce_l2_factor

    l1b_file = l1b.get_nc_filename()

    rp = RetrackerProcessor(
        l1b_source=[l1b_file],
        l1b_data_vars=data_vars_l1b,
        rp_sets=rp_sets,
        retrack_sets=retrack_sets,
        fitting_sets=fitting_sets,
        sensor_sets=sensor_sets,
        wf_sets=wf_sets,
        # log_level=logging.DEBUG,
    )
    rp.process()

    # ref datasets
    l2_ref = l2(n_offset=0, n_inds=0)
    l2_mask = np.in1d(l2_ref["record_inds"], rp.output_l2.record_ind.values)
    l2_ref = l2(n_offset=l2_mask.nonzero()[0][0], n_inds=len(l2_mask.nonzero()[0]))

    fig, rmse_swh, rmse_ssh = plot_l2_results_vs_ref(
        rp.output_l2,
        l2_ref,
        fig_title=f"{rp_sets.retracker_basetype}-{preset} ({l1b_src_type.value})",
    )

    fig.show()

    # assert rmse_swh < max_rmse_swh_m
    # assert rmse_ssh < max_rmse_ssh_m



#!/usr/bin/env python

"""The setup script."""

import sys
from setuptools import setup, find_packages, Extension
import re

with open("README.md") as readme_file:
    readme = readme_file.read()

# with open("HISTORY.rst") as history_file:
#     history = history_file.read()

requirements = []
with open("requirements.txt") as f:
    requirements = f.read().splitlines()
    requirements = [
        r
        for r in requirements
        if re.split("[^a-zA-Z0-9]", r)[0]
        not in [
            "python",
            "watchdog",
            "pip",
            "bump2version",
            "tox",
            "flake8",
            "imageio",
            "pytest",
            "wheel",
            "coverage",
            "twine",
        ]
    ]

test_requirements = ["pytest>=3"]

extensions = [
    Extension("pysamosa.model_helpers", ["pysamosa/model_helpers.pyx"]),
]

setup(
    author="Florian Schlembach",
    author_email="florian.schlembach@tum.de",
    python_requires=">=3.8",
    classifiers=[
        "Development Status :: 4 - Beta",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
        "Natural Language :: English",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
    ],
    description="PySAMOSA is a software framework for processing open ocean and coastal waveforms from SAR satellite altimetry to measure sea surface heights, wave heights, and wind speed for the oceans and inland water bodies. Satellite altimetry is a space-borne remote sensing technique used for Earth observation.",
    install_requires=requirements,
    license="GNU General Public License v3",
    long_description=readme,
    long_description_content_type="text/markdown",
    include_package_data=True,
    keywords="pysamosa satellite altimetry retracking samosa+ samosa coastal open ocean sentinel esa"
    "sar altimetry ff-sar fully focused",
    name="pysamosa",
    packages=find_packages(include=["pysamosa", "pysamosa.*"]),
    test_suite="tests",
    tests_require=test_requirements,
    ext_modules=extensions,
    url="https://github.com/floschl/pysamosa",
    version="0.2.10",
    zip_safe=False,
    setup_requires=[],
)



import os
import logging
import numpy as np

import matplotlib as mpl
import matplotlib.pyplot as plt
from matplotlib.backends.backend_pdf import PdfPages
from matplotlib.backends.backend_pgf import FigureCanvasPgf

from scipy import signal

default_width_in = 5.79  # in inches
default_ratio = 0.5 + np.sqrt(5) / 2  # 1.618, golden ratio
default_figsize_in = [default_width_in, default_width_in / default_ratio]


def gen_first_true():
    yield True
    while True:
        yield False


def consecutive_regions_from_ind_list(data, stepsize=1):
    regions = np.split(data, np.where(np.diff(data) != stepsize)[0] + 1)
    return regions if len(data) else []


def plot_single_retrack_result(
    *,
    ax,
    wf_meas,
    res_fit,
    fg_epoch,
    retrack_sets,
    l1b_data_single,
    fontsize_labels=None,
    legend_kwargs=None,
    subplot_name=None,
    subplot_name_kw=None,
    xlabel=None,
    ylabel=None,
    do_plot_second_halve_only=False,
):
    _ax = ax

    fontsize_labels = 8 if fontsize_labels is not None else fontsize_labels
    legend_kwargs = {} if legend_kwargs is None else legend_kwargs
    subplot_name = "" if subplot_name is None else subplot_name
    subplot_name_kw = {} if subplot_name_kw is None else subplot_name_kw

    plot_data = [
        {"name": r"$\mathbf{w}_\mathrm{r}$", "wf": wf_meas},
        {"name": "$\\mathbf{w}_\\mathrm{SAM2}$", "wf": res_fit["wf_opt"]},
        # {'name': '$\mathbf{w}_{\mathrm{SAM2},i}$\n$(\mathrm{SWH}_\mathrm{aim\_mask}=$' + f'{res_fit["swh"]:.2f}m)', 'wf': res_fit['wf_opt']},
    ]

    start_ind = len(plot_data[0]["wf"]) // 2 if do_plot_second_halve_only else 0

    for d in plot_data:
        _ax.plot(d["wf"][start_ind:], linewidth=0.7, label=d["name"])

    # plot initial first-guess epoch
    fg_epoch_style = {
        "linestyle": "--",
        "linewidth": 1,
        "color": "grey",
        "label": "$k_\\mathrm{DFGE}$",
    }
    fg_epoch -= start_ind
    _ax.axvline(fg_epoch, **fg_epoch_style)

    if "interference_inds" in res_fit:
        distorted_regions = [
            (dr - start_ind)
            for dr in consecutive_regions_from_ind_list(res_fit["interference_inds"])
        ]
        show_legend_entry_it = gen_first_true()
        for br in distorted_regions:
            _ax.axvspan(
                br[0],
                br[-1],
                alpha=0.5,
                color="red",
                label="$\\mathbf{k}_\\mathrm{inf}$"
                if next(show_legend_entry_it)
                else "",
            )

    if "interference_mask" in res_fit and not np.allclose(
        res_fit["interference_mask"], 1.0
    ):
        _ax.plot(
            res_fit["interference_mask"][start_ind:],
            linewidth=0.7,
            label="$\\mathbf{w}_\\mathrm{IR}$",
            color="lime",
        )

    # if res_fit['le_inds'] is not None:
    #     le_inds = res_fit['le_inds']
    #     _ax.axvspan(le_inds[0], le_inds[-1], alpha=0.5, color='lightblue')

    if retrack_sets.subwaveform_mode:
        _ax.axvspan(
            res_fit["max_le_gate"] + retrack_sets.subwaveform_n_gates_after_le,
            len(res_fit["wf"]),
            alpha=0.5,
            color="gray",
        )

    _ax.legend(**legend_kwargs)

    # info box
    fontsize_textboxes = 6
    text_str_elems = [
        f"dist2coast: {l1b_data_single['dist2coast']:.0f} km",
    ]
    _ax.text(
        x=0.95,
        y=0.05,
        s="\n".join(text_str_elems),
        ha="right",
        va="center",
        transform=_ax.transAxes,
        fontsize=fontsize_textboxes,
        bbox=dict(pad=0.5, fc="white", alpha=0.8),
    )

    _ax.set_ylim(bottom=-0.1, top=1.1)
    if xlabel:
        _ax.set_xlabel(xlabel, fontsize=fontsize_labels)
    if ylabel:
        _ax.set_ylabel(ylabel, fontsize=fontsize_labels)
    if subplot_name:
        _ax.annotate(subplot_name, **subplot_name_kw)
    _ax.grid()


def save_figs_to_pdf(figs, pdf_filepath):
    """Saves objects to disk using pickle."""
    os.umask(000)
    os.makedirs(os.path.dirname(pdf_filepath), exist_ok=True)

    pp = PdfPages(pdf_filepath)

    figs = figs if isinstance(figs, list) else [figs]

    for f in figs:
        pp.savefig(f, box_inches="tight", pad_inches=0)

    pp.close()

    logging.info("Figure saved to pdf file {}".format(pdf_filepath))


def set_pgf_mode():
    # register pdf extension with FigureCanvasPgf backend
    # https://stackoverflow.com/questions/9169052/partial-coloring-of-text-in-matplotlib/42768093#42768093
    mpl.backend_bases.register_backend("pgf", FigureCanvasPgf)
    mpl.backend_bases.register_backend("pdf", FigureCanvasPgf)

    mpl.rcParams.update(
        {
            "pgf.rcfonts": False,
            "text.usetex": True,
            "pgf.texsystem": "pdflatex",
            "font.family": "serif",
            "font.serif": [],  # use latex default serif font
            "font.sans-serif": [],
            "font.monospace": [],
            "figure.figsize": default_figsize_in,
            "figure.constrained_layout.use": True,
            "pgf.preamble": r"\usepackage[utf8x]{inputenc}\usepackage[T1]{fontenc}",
        }
    )


def plot_l2_results_vs_ref(l2, l2_ref, cog_corr=0.0, fig_title=None):
    fontsize_labels, fontsize_legend = 7, 7

    fig, axs_data = plt.subplots(2, 1)

    # lat = np.degrees(l2.latitude)
    lat = l2.latitude

    # SWH
    swh_diff = l2.swh - l2_ref["swh"]
    rmse_swh = float(np.sqrt(np.mean((swh_diff) ** 2)).values)
    median_bias_swh = np.median(swh_diff)
    lh_retrack = axs_data[0].plot(
        lat,
        l2.swh,
        linewidth=1.0,
        label=f"swh_retrack (std={np.nanstd(l2.swh.values):.2f},"
        f"median={np.nanmedian(l2.swh.values):.2f}, bias_eum={median_bias_swh:.4f}m, rmse_eum={rmse_swh:.4f}m)",
        zorder=2,
    )
    axs_data[0].plot(
        lat,
        l2_ref["swh"],
        linewidth=0.8,
        label=f'swh_ref (std={np.nanstd(l2_ref["swh"]):.2f}, median={np.nanmedian(l2_ref["swh"]):.2f})',
    )

    # SWH quality flag
    axs_data[0].plot(
        lat,
        l2.swh_qual,
        color=lh_retrack[0].get_color(),
        linewidth=1.0,
        linestyle="--",
        label="swh_qual",
    )

    # uncorrected SSH
    ssh_uncorr_rt = l2.altitude - l2.range
    ssh_uncorr_ref = l2_ref["alt_m"] - l2_ref["range"] + cog_corr
    ssh_diff = ssh_uncorr_rt - ssh_uncorr_ref

    rmse_ssh = float(np.sqrt(np.mean((ssh_diff) ** 2)).values)
    median_bias_ssh = np.median(ssh_diff)

    def nanstd_detrend(alt, range):
        non_nan_mask = ~(np.isnan(alt) | np.isnan(range))
        return (
            np.nanstd(signal.detrend(alt[non_nan_mask] - range[non_nan_mask]))
            if any(non_nan_mask)
            else np.nan
        )

    axs_data[1].plot(
        lat,
        ssh_uncorr_rt,
        linewidth=1.0,
        label=f"l2 (std_detrend={nanstd_detrend(l2.altitude, l2.range):.2f}m, median_bias={median_bias_ssh:.4f}m, RMSE={rmse_ssh:.4f}m)",
    )
    axs_data[1].plot(
        lat,
        ssh_uncorr_ref,
        linewidth=1.0,
        label=f'l2_ref (std_detrend={nanstd_detrend(l2_ref["alt_m"], l2_ref["range"]):.2f}m)',
    )

    # plot settings
    axs_data[0].legend(fontsize=fontsize_legend, loc="lower left")
    axs_data[0].set_ylabel("SWH [m]", fontsize=fontsize_labels)
    axs_data[0].grid()
    axs_data[0].tick_params(axis="both", which="major", labelsize=fontsize_labels)

    # axs_data[1].set_ylim([np.nanmin(l2.altitude - l2.range), np.nanmax(l2.altitude - l2.range)])
    axs_data[1].legend(fontsize=fontsize_legend, loc="lower left")
    axs_data[1].set_ylabel("uncorrected SSH [m]", fontsize=fontsize_labels)
    axs_data[1].grid()
    axs_data[1].tick_params(axis="both", which="major", labelsize=fontsize_labels)
    axs_data[1].set_ylim(np.min(l2.altitude - l2.range), np.max(l2.altitude - l2.range))

    if fig_title:
        fig.suptitle(fig_title, fontsize=fontsize_labels)

    return fig, rmse_swh, rmse_ssh



import os
import pickle
import numpy as np
import cython


def load_samosa_luts():
    with open(os.path.dirname(__file__) + f'/luts_samosa.pickle', 'rb') as handle:
        return pickle.load(handle)

from libc.math cimport sqrt, M_PI

@cython.cdivision(True)
@cython.boundscheck(False)
cpdef get_f_from_lut(double[:] xi, double[:,:] f_lut, int degree):
    ''' Reads the F0/F1 values from the LUT as described in DPM doc v2.5.2 4.2.3 SL7

    :param xi: the value to look up
    :param f_lut: either LUT_F0 or LUT_F1
    :param degree: 0: for LUT_F0, 1: LUT_F1
    :return: the looked up value
    '''
    farr = np.zeros(xi.shape[0])
    cdef double[:] farr_view = farr

    set_lut_val(xi, f_lut, degree, farr_view)

    return farr

@cython.cdivision(True)
@cython.boundscheck(False)
cdef double[:] set_lut_val(double[:] xi, double[:,:] f_lut, int degree, double[:] farr_view) nogil:

    cdef double[:] f_lut_x = f_lut[:, 0]
    cdef double[:] f_lut_y = f_lut[:, 1]

    cdef double f = 0.0
    cdef int j = 0

    for i in range(xi.shape[0]):
        if xi[i] >= f_lut_x[0] and xi[i] <= f_lut_x[-1]:
            j = <int>((f_lut_x.shape[0] - 1) * (xi[i] - f_lut_x[0]) / (f_lut_x[-1] - f_lut_x[0]))
            f = (xi[i] - f_lut_x[j]) * (f_lut_y[j + 1] - f_lut_y[j]) / (f_lut_x[j + 1] - f_lut_x[j]) + f_lut_y[j]
        elif xi[i] < f_lut_x[0]:
            f = 0.0
        else:
            if degree == 0:
                f = 1.0 / 2.0 * sqrt(2.0 * M_PI / xi[i])
            elif degree == 1:
                f = 1.0 / 4.0 * sqrt(2.0 * M_PI / (xi[i] ** 3))

        farr_view[i] = f

    return farr_view


import numpy as np

from pysamosa.model import get_region_max


def detect_leading_edge(
    wf, fg_epoch=None, normalise_wf_by_fg_region=None, append_first_falling_edge=False
):
    """Detects the leading edge in a power return echo waveform

    :raises ValueError: wrong input arguments
    :raises RuntimeError: Leading edge detection failed
    :param wf: the waveform in samples
    :param fg_epoch: the first-guess epoch, if None is given the maximum of the waveform is chosen
    :param normalise_wf_by_fg_region: number of samples in gates around the fg_epoch the waveform is normalised
    :return: the indices of the gates for which the leading edge is detected, type np.ndarray
    """
    if fg_epoch is not None and normalise_wf_by_fg_region is None:
        raise ValueError("Error in leading edge detection. ")

    if fg_epoch and normalise_wf_by_fg_region:
        wf = wf / get_region_max(
            wf=wf, region_center=fg_epoch, n_before_after=normalise_wf_by_fg_region
        )
    else:
        wf = wf / np.max(wf)

    fg_epoch = np.argmax(wf) if fg_epoch is None else fg_epoch

    # get the numerical gradient of the waveform and all rising edges above a
    # certain threshold
    def get_rising_falling_edge(_wf, falling_edges=False):
        d = np.diff(_wf, prepend=[0])
        le_ind_mask = (d < 0.01) if falling_edges else (d > 0.01)

        # get all consecutive rising edges
        max_stepsize = 1
        data = le_ind_mask.nonzero()[0]
        return np.split(data, np.where(np.diff(data) > max_stepsize)[0] + 1)

    # remove indices from edges that are after fg_epoch
    rising_edges = get_rising_falling_edge(wf)

    rising_edges_shortened = [
        list(r)
        for r in rising_edges
        if (fg_epoch - 2 * normalise_wf_by_fg_region)
        <= r[-1]
        <= (fg_epoch + normalise_wf_by_fg_region)
    ]
    rising_edges_shortened = list(
        filter(None, rising_edges_shortened)
    )  # remove empty list

    # selected_le_gates = sorted(rising_edges_left_fg_shortend, key=len,
    # reverse=True)[0]  # sort rising edges by width
    selected_le_gates = sorted(
        rising_edges_shortened, key=lambda e: wf[e[-1]] - wf[e[0]], reverse=True
    )  # sort rising edges by height

    # fill potential gaps in series
    if len(selected_le_gates) and len(selected_le_gates[0]):
        selected_le_gates = np.arange(
            selected_le_gates[0][0], selected_le_gates[0][-1] + 1
        )
    else:
        raise RuntimeError("Leading edge could not be detected. ")

    # detect trailing edge after leading edge
    if append_first_falling_edge:
        falling_edges = get_rising_falling_edge(wf, falling_edges=True)
        falling_edge_after_le = [
            fe for fe in falling_edges if (fe[0] - 1) == selected_le_gates[-1]
        ][0]

        selected_le_gates = np.concatenate((selected_le_gates, falling_edge_after_le))

    return selected_le_gates


# class LeadingEdgeDetector():
#
#     def __init__(self):
#         self.init = True
#
#     def __call__(self, *args, **kwargs):
#         wf =



import numpy as np
import logging

# import imageio.v3 as iio
import imageio.v2 as iio

from scipy.io import savemat


def dict_remove_None(d):
    return dict(filter(lambda item: item[1] is not None, dict(d).items()))


def replace_none(test_dict):
    # checking for dictionary and replacing if None
    if isinstance(test_dict, dict):
        for key in test_dict:
            if test_dict[key] is None:
                test_dict[key] = ""
            else:
                replace_none(test_dict[key])

    # checking for list, and testing for each value
    elif isinstance(test_dict, list):
        for val in test_dict:
            replace_none(val)


class SettingsDumper:
    def __init__(
        self, *, wf_sets, model_sets, sensor_sets, dest_mat_file, dest_gif_file
    ):
        self.dest_mat_file = dest_mat_file
        self.dest_gif_file = dest_gif_file
        self.l1b_nc_filename = None
        self.meas = []
        self.figs = []

        self.all_data = {
            "wf_sets": wf_sets.dict(),
            "model_sets": model_sets.dict(),
            "sensor_sets": sensor_sets.dict(),
            "meas": self.meas,
        }

    def add_retrack_entry(
        self, *, l1b_data_single, model_params, record_ind, res_fit, l2_data_single
    ):
        self.meas.append(
            {
                "record_ind": record_ind,
                "model_params": dict_remove_None(model_params),
                "l1b_data_single": dict_remove_None(l1b_data_single),
                "l2_data_single": l2_data_single,
                "res_fit": dict_remove_None(res_fit),
            }
        )

    def add_fig(self, f):
        self.figs.append(f)

    def write_out_mat(self):
        self.all_data["l1b_nc_filename"] = str(self.l1b_nc_filename)

        replace_none(self.all_data)

        savemat(str(self.dest_mat_file), self.all_data)
        logging.info(f"Writing retracked data to file {self.dest_mat_file}...")

    def export_gif(self):
        # save tmp pngs
        tmp_imgs = []
        for i, f in enumerate(self.figs):
            p = self.dest_gif_file.parent / f"tmp_img_{i}.png"
            f.savefig(p, dpi=200)
            tmp_imgs.append(p)

        if tmp_imgs:
            frames = np.stack([iio.imread(f) for f in tmp_imgs], axis=0)
            # iio.imwrite(self.dest_gif_file, frames, duration=0.2, loop=0)
            iio.mimsave(self.dest_gif_file, frames, duration=500, loop=0)



import numbers
from textwrap import wrap

import matplotlib.pyplot as plt
import numpy as np

from pysamosa.utils import consecutive_regions_from_ind_list

from pysamosa.common_types import (
    RetrackerProcessorSettings,
    WaveformSettings,
    ModelSettings,
    RetrackerSettings,
    SensorSettings,
    L1bSourceType,
    SENSOR_SETS_DEFAULT_S3,
)
from pysamosa.conf_params import CONST_C
from pysamosa.data_access import get_model_param_obj_from_l1b_data, get_subset_dataset
from pysamosa.model import SamosaModel
from pysamosa.retracker import calc_misfit, SamosaRetracker
from pysamosa.retracker_helpers import get_dynamic_first_guess_epochs
from pysamosa.rip import RipAnalyser, RIPParameters


def plot_retrack_result(
    l1b_data,
    l2_data,
    res_fit,
    model_params,
    *,
    l1b_srctype=L1bSourceType,
    sensor_sets: SensorSettings = None,
    wf_sets: WaveformSettings = None,
    model_sets: ModelSettings = None,
    retrack_sets: RetrackerSettings = None,
    ax=None,
    title=None,
    legend_kwargs=None,
    legend_wrap_len=None,
    show_l2_model=True,
    show_second_halve=False,
    show_leading_edge=False,
):
    if ax:
        _ax = ax
    else:
        _ax = plt.gca()

    fg_epoch = (
        l1b_data["dynamic_fg_epoch"]
        if "dynamic_fg_epoch" in l1b_data
        else np.argmax(l1b_data["wf"])
    )
    if wf_sets is not None:
        fg_epoch *= wf_sets.internal_oversampling_factor

    wf_len = len(res_fit["wf"])
    wf_sets = (
        wf_sets
        if wf_sets is not None
        else WaveformSettings.get_default_src_type(l1b_srctype)
    )

    if retrack_sets and retrack_sets.normalise_wf_by_fg_region:
        wf_meas = res_fit["wf"] / SamosaRetracker.get_wf_max(
            res_fit["wf"], retrack_sets=retrack_sets, fg_epoch=fg_epoch
        )
    else:
        wf_meas = res_fit["wf"] / np.max(res_fit["wf"])

    sm = SamosaModel(sensor_sets=sensor_sets, model_sets=model_sets, wf_sets=wf_sets)

    Pu_W_l2 = l2_data["Pu_W"] if "Pu_W" in l2_data else np.nan
    Pu_W_fit = res_fit["Pu_denorm"]
    l2_epoch_ns = (
        l2_data["epoch_ns"]
        if "epoch_ns" in l2_data and not np.isnan(l2_data["epoch_ns"])
        else np.nan
    )
    try:
        wf_l2_model = sm.get_waveform_multilook(
            Pu=Pu_W_l2, Hs=l2_data["swh"], t0_ns=l2_epoch_ns, model_params=model_params
        )
    except Exception:
        wf_l2_model = np.zeros(wf_len)
    wf_l2_model = wf_l2_model / np.nanmax(wf_l2_model)
    if show_l2_model:
        wf_l2_model_l2norm = np.sum((wf_meas - wf_l2_model) ** 2)
        misfit_l2_model = calc_misfit(wf_meas, wf_l2_model)
    # ratio_Pu_l2_to_fit = Pu_W_l2 / Pu_W_fit
    # ratio_Pu_l2_to_wf = Pu_W_l2 / np.max(l1b_data['wf'])

    # data to be plotted
    range_l2 = l2_data["range"]
    tracker_range_m = (
        l2_data["tracker_range_m"]
        if "tracker_range_m" in l2_data
        else l1b_data["tracker_range_m"]
    )
    range_fit = tracker_range_m + (float(res_fit["epoch_ns"])) * 1e-9 * (CONST_C / 2)
    ssh_unc_l2 = l2_data["alt_m"] - range_l2
    ssh_unc_fit = l1b_data["alt_m"] - range_fit
    n_iter = l2_data["n_iter"] if "n_iter" in l2_data else np.nan
    plot_data = [
        {
            "name": "y_l2",
            "wf": wf_meas,
            "misfit": l2_data.get("misfit", np.nan),
            "l2norm": np.nan,
            "n_iter": n_iter,
            "swh": l2_data["swh"],
            "epoch_ns": float(l2_epoch_ns),
            "range": range_l2,
            "ssh_unc": ssh_unc_l2,
            "Pu_W": Pu_W_l2,
            "nu": np.nan,
        },
        {
            "name": "y_retrack",
            "wf": res_fit["wf_opt"],
            "misfit": res_fit["misfit"],
            "misfit_selective": res_fit["misfit_selective"],
            "misfit_le": res_fit["misfit_le"] if "misfit_le" in res_fit else np.nan,
            "l2norm": res_fit["l2norm"],
            "n_iter": res_fit["n_iter"],
            "swh": res_fit["swh"],
            "epoch_ns": res_fit["epoch_ns"],
            "range": range_fit,
            "ssh_unc": ssh_unc_fit,
            "Pu_W": Pu_W_fit,
            "nu": res_fit["nu"],
        },
    ]

    if show_l2_model:
        plot_data.append(
            {
                "name": "y_meas_model",
                "wf": wf_l2_model,
                "misfit": misfit_l2_model,
                "l2norm": wf_l2_model_l2norm,
                "n_iter": np.nan,
                "swh": l2_data["swh"],
                "epoch_ns": l2_epoch_ns,
                "range": range_l2,
                "ssh_unc": ssh_unc_l2,
                "Pu_W": Pu_W_l2,
                "nu": np.nan,
            }
        )

    legend_len = 90 if legend_wrap_len is None else legend_wrap_len
    for d in plot_data:
        _ax.plot(
            d["wf"],
            linewidth=0.7,
            label="\n".join(
                wrap(
                    f"{d['name']}, misfit={d['misfit']:.2f},"
                    f" misfit_selective={d['misfit_selective'] if 'misfit_selective' in d else np.nan:.2f}, "
                    f" misfit={d['misfit'] if 'misfit' in d else np.nan:.2f}, "
                    # f" misfit_le={(d['misfit_le'] if 'misfit_le' in d else np.nan):.2f}"
                    f" n_iter={d['n_iter']}, "
                    f"SWH={d['swh']:.3f}m, "
                    f"epoch_ns={d['epoch_ns']:.3f}ns,"
                    f"range={d['range']:.3f}m,"
                    f"ssh_unc={d['ssh_unc']:.3f}m,"  # noqa
                    # f"Pu={d['Pu_W']:.2e}W"
                    # f"nu={d['nu']:.2e}"
                    ,
                    width=legend_len,
                )
            ),
        )

    # plot initial first-guess epoch
    fg_epoch_style = {
        "linestyle": "--",
        "linewidth": 1,
        "color": "grey",
        "label": "Dynamic First-Guess Epoch (DFGE)",
    }
    _ax.axvline(fg_epoch, **fg_epoch_style)

    if "interference_inds" in res_fit:
        distorted_regions = consecutive_regions_from_ind_list(
            res_fit["interference_inds"]
        )
        for br in distorted_regions:
            _ax.axvspan(br[0], br[-1], alpha=0.5, color="red")

    if "interference_mask" in res_fit and not np.allclose(
        res_fit["interference_mask"], 1.0
    ):
        _ax.plot(
            res_fit["interference_mask"],
            linewidth=0.7,
            label="interference reference waveform",
        )

    if show_leading_edge and res_fit["le_inds"] is not None:
        le_inds = res_fit["le_inds"]
        _ax.axvspan(
            le_inds[0], le_inds[-1], alpha=0.5, color="lightblue", label="leading edge"
        )

    if retrack_sets is not None and retrack_sets.subwaveform_mode:
        _ax.axvspan(
            res_fit["max_le_gate"] + retrack_sets.subwaveform_n_gates_after_le,
            wf_len,
            alpha=0.5,
            color="gray",
        )

    # labelling
    fontsize_labels = 7
    fontsize_textbox = 6
    _ax.set_title("" if title is None else title, fontsize=fontsize_labels)
    legend_kwargs_default = {
        "loc": "lower left",
        "bbox_to_anchor": (0, 1.03, 1, 0),
        "mode": "expand",
        **{
            "prop": {"size": fontsize_textbox},
            "labelspacing": 0.50,
            "borderaxespad": 0.2,
        },
    }
    if legend_kwargs is not None:
        legend_kwargs_default = {**legend_kwargs_default, **legend_kwargs}
    _ax.legend(**legend_kwargs_default)

    # info box
    fontsize_textboxes = 6
    text_str_elems = [
        f"dist2coast: {l1b_data['dist2coast']:.0f} km",
    ]
    _ax.text(
        x=0.95,
        y=0.95,
        s="\n".join(text_str_elems),
        ha="right",
        va="center",
        transform=_ax.transAxes,
        fontsize=fontsize_textboxes,
        bbox=dict(pad=0.5, fc="white", alpha=0.8),
    )

    if show_second_halve:
        _ax.set_xlim(left=wf_len // 2, right=wf_len)
    _ax.set_ylim(bottom=-0.1, top=1.1)
    _ax.set_xlabel("Range gate k", fontsize=fontsize_labels)
    _ax.set_ylabel("Normalised power", fontsize=fontsize_labels)
    _ax.grid()

    if not ax:
        plt.show()


def plot_rip_result(
    l1b_data_single: dict, ripa: RipAnalyser, rip_params: RIPParameters, ax=None
):
    if ax is not None:
        _ax = ax
    else:
        _ax = plt.gca()

    _ax.plot(ripa.doppler_beam_inds, ripa.rip_wf_norm, label="measured RIP")
    _ax.plot(
        ripa.doppler_beam_inds, rip_params.rip_az_fitted, "+", label="fitted gaussian"
    )
    _ax.axvline(
        rip_params.pitch_mispoint_look, color="red", linestyle="dashed", linewidth=0.5
    )
    _ax.legend(fontsize=7, loc="upper left")

    _ax.plot(
        rip_params.halfpower_looks,
        0.5 * rip_params.amplitude_fitted_norm * np.ones(2),
        color="red",
        linestyle="dashed",
        linewidth=0.5,
        label="3db aperture",
    )

    # labelling
    fontsize = 7
    fontsize_textboxes = 6

    # RIP parameter text box
    text_str = "\n".join(
        [
            f"{k} = {v:.3e}"
            for k, v in rip_params.dict().items()
            if isinstance(v, numbers.Number)
        ]
    )
    _ax.text(
        x=0.98,
        y=0.05,
        s=text_str,
        ha="right",
        va="bottom",
        transform=_ax.transAxes,
        fontsize=fontsize_textboxes,
        bbox=dict(pad=0.4, fc="white", alpha=0.8),
    )

    # measured params
    l1b_params = {"pitch_mispoint_rad_instr": l1b_data_single["ksix_rad"]}
    text_str_elems = [*[f"{k} = {v:.3e}" for k, v in l1b_params.items()]]
    _ax.text(
        x=0.98,
        y=0.95,
        s="\n".join(text_str_elems),
        ha="right",
        va="top",
        transform=_ax.transAxes,
        fontsize=fontsize_textboxes,
        bbox=dict(pad=0.4, fc="white", alpha=0.8),
    )

    _ax.set_xlabel("integer looks", fontsize=fontsize)
    _ax.set_ylabel("normalised power", fontsize=fontsize)

    if not ax:
        plt.show()


def get_l1b_data_single_from_l1b_fixture(
    *, l1b_fixture, ind, rp_sets: RetrackerProcessorSettings, file_id
):
    n_inds_total = rp_sets.dynamic_fg_epoch_n_adjacent_meas
    n_pre = n_inds_total // 2

    l1b_data = l1b_fixture(n_offset=ind - n_pre, n_inds=n_inds_total, file_id=file_id)
    nc_file = l1b_fixture.get_nc_filename(file_id)
    l1b_data["dynamic_fg_epoch"] = get_dynamic_first_guess_epochs(
        wfs=l1b_data["wf"],
        tracker_range=l1b_data["tracker_range_m"],
        alt_m=l1b_data["alt_m"],
        bu_bw_Hz=SENSOR_SETS_DEFAULT_S3.B_r_Hz,
        fg_epoch_adjacent_meas=rp_sets.dynamic_fg_epoch_n_adjacent_meas,
    )
    l1b_data_single = get_subset_dataset(l1b_data, ind_offset=n_pre)
    model_params_single = get_model_param_obj_from_l1b_data(l1b_data, ind=n_pre)

    return l1b_data_single, model_params_single, nc_file



# taken from SAMOSA DPM Doc 2.5.2

# S-3 Ku Band SAR Re-tracker Configuration File
CONST_A = 6378137  # Equatorial radius [m]
CONST_B = 6356752.3142  # Polar radius [m]
# CONST_B = 6.356752314245179e+06  # Polar radius [m]
CONST_C = 299792458  # speed of light [m/s]
CONST_F = 0.003352810664747  # flattening coefficient f from WGS84
